"""sonusai tplot

usage: tplot [-hv] [-r ROOT] [-c CONFIG] [-t TARGET] [-m MODEL] [-f FEATURE] [--sedthr SEDTHR]

options:
   -h, --help
   -v, --verbose            Be verbose.
   -r, --root ROOT          Root of a gentcst tree.
   -c, --config CONFIG      YAML config file.
   -t, --target TARGET      Target .wav or .txt, else use target section in CONFIG.
   -m, --model MODEL        Aaware ONNX model file.
   -f, --feature FEATURE    Feature override, e.g., gfh64to128.
   --sedthr SEDTHR          SED thresholds override, e.g., '[-24, -27, -30]'.

   CONFIG is used to override the default configuration, thus it does not need to be a complete
   genmixdb config parameter set, only desired overrides like truth function and truth config.

   A TARGET .txt file supports the same format as SonusAI genmixdb which is a list of regexes
   for finding sound files.

   ROOT, CONFIG, and TARGET are optional, but at least one of them must be specified.
   Behavior is as follows in order of precedence:

    1)  If ROOT is provided, use the given directory as the root directory of
        a gentcst tree. The config will be generated from the top down to the current leaf.
    2)  If CONFIG is provided, use the given YAML file to override the config.
    3)  If TARGET is provided, use the given .wav or .txt file to override the target
        section of the config.
    4)  If FEATURE and/or SEDTHR are provided, use them to override the config.

   If TARGET is given without a CONFIG, look for a corresponding YAML file to use as a
   config. For example,

        sonusai tplot -t 4-esc50-dogbark.txt

   will first look for '4-esc50-dogbark.yml' and then for 'config.yml'. If neither
   candidate was found and ROOT was also not specified then give an error. If
   ROOT was specified then use the hierarchical config.

   A multi-page plot TARGET-tplot.pdf or CONFIG-tplot.pdf is generated.

"""

from sonusai import logger


def tplot(root: str = None,
          config_file: str = None,
          target_file: str = None,
          model_file: str = None,
          feature: str = None,
          thresholds: str = None,
          verbose: bool = False) -> None:
    import json
    import os

    import matplotlib.pyplot as plt
    import numpy as np
    import yaml
    from matplotlib.backends.backend_pdf import PdfPages
    from pyaaware import Predict

    import sonusai
    from sonusai import SonusAIError
    from sonusai import logger
    from sonusai.genft import genft
    from sonusai.genmix import genmix
    from sonusai.genmixdb import genmixdb
    from sonusai.mixture import get_default_config
    from sonusai.mixture import update_config_from_file
    from sonusai.mixture import update_config_from_hierarchy
    from sonusai.utils import int16_to_float

    figsize = (11, 8.5)

    if config_file is None and target_file is None and root is None:
        raise SonusAIError('Error: No config provided.')

    target_path = ''
    config = get_default_config()

    if root is not None:
        leaf = os.getcwd()
        target_path = os.path.basename(leaf)
        config = update_config_from_hierarchy(root=root, leaf=leaf, config=config)
        logger.debug(f'Hierarchical config for {leaf} using {root} as the root')

    if config_file is not None:
        target_path = config_file
        config = update_config_from_file(config_file, config)
        logger.debug(f'Config from {config_file}')

    if target_file is not None:
        target_path = target_file
        if not os.path.exists(target_file):
            raise SonusAIError(f'{target_file} does not exist.')

        config['targets'] = [{'name': target_file}]
        logger.debug(f'Targets from {target_file}')

        if config_file is None:
            target_config = f'{os.path.splitext(target_file)[0]}.yml'
            if os.path.exists(target_config):
                config = update_config_from_file(target_config, config)
                logger.debug(f'Config from {target_config}')
            else:
                target_config = 'config.yml'
                if os.path.exists(target_config):
                    config = update_config_from_file(target_config, config)
                    logger.debug(f'Config from {target_config}')
                elif root is None:
                    raise SonusAIError(f'Error: Could not determine config for {target_file}.')

    file_base = os.path.splitext(target_path)[0]
    file_name = os.path.split(file_base)[1]
    output = f'./{file_name}-tplot.pdf'

    if len(config['truth_settings']) > 1:
        raise SonusAIError(f'Number of truth_settings is more than one. This is not supported yet.')

    if thresholds is not None:
        thresholds = json.loads(thresholds)
        if thresholds != config['truth_settings'][0]['config']['thresholds']:
            config['truth_settings'][0]['config']['thresholds'] = thresholds
            logger.debug(f'Override SED thresholds with {thresholds}')

    if feature is not None:
        if feature != config['feature']:
            config['feature'] = feature
            logger.debug(f'Override feature with {feature}')

    logger.debug('')

    # Run genmixdb, genmix, and genft
    mixdb = genmixdb(config=config, logging=verbose)
    num_mixtures = len(mixdb.mixtures)  # Number of mixtures
    logger.info(f'Generating data for {num_mixtures} mixture plots')
    mixture, _, target, noise, _, mixdb_out = genmix(mixdb=mixdb,
                                                     mixids='*',
                                                     compute_segsnr=False,
                                                     compute_truth=False, )

    feature_data, truth_f, _, _ = genft(mixdb=mixdb,
                                        mixids='*',
                                        compute_segsnr=False)

    # If model provided, read and check it
    predict_data = np.zeros(0)
    if model_file is not None:
        import onnx

        # model is an in-memory ModelProto
        model = onnx.load(model_file)
        if len(model.metadata_props) < 5:
            logger.warn('Model metadata indicates this is not an Aaware model, ignoring.')
        else:
            if model.metadata_props[4].key != 'feature':
                logger.warn('Model metadata does not have Aaware feature, ignoring.')
            else:
                model_feature = model.metadata_props[4].value
                logger.debug(f'Model feature is {model_feature}')
                # TODO Check and read other params, flatten, addch, etc.
                logger.info(f'Running prediction with {model_file}')
                predict = Predict(model_file)
                predict_data = predict.execute(feature_data)

    num_samples = mixture.shape[0]  # Total number of samples over all mixtures
    num_features = feature_data.shape[0]  # Total number of feature frames over all mixtures

    logger.info(f'Plotting {num_mixtures} target-truth results to {output}')
    pdf = PdfPages(f'{output}')
    for m in range(num_mixtures):
        mixture_begin = mixdb_out['mixtures'][m]['i_sample_offset']
        feature_begin = mixdb_out['mixtures'][m]['o_frame_offset']
        # For each target/mixture, get index endpoint (mix,noise,target sample,...)
        if m == num_mixtures - 1:
            mixture_end = num_samples
            feature_end = num_features
        else:
            mixture_end = mixdb_out['mixtures'][m + 1]['i_sample_offset']
            # out frames are possibly decimated/stride-combined
            feature_end = mixdb_out['mixtures'][m + 1]['o_frame_offset']

        # Trim waveform data
        mixture_trimmed = int16_to_float(mixture[mixture_begin:mixture_end])
        target_trimmed = int16_to_float(target[mixture_begin:mixture_end])

        # Provide subset defined by mixture using truth_index array of indices (multichannel truth)
        target_record = mixdb_out['targets'][mixdb_out['mixtures'][m]['target_file_index']]
        truth_settings = target_record['truth_settings']
        if len(truth_settings) > 1:
            raise SonusAIError(f'Number of truth_settings is more than one. This is not supported yet.')

        truth_index = np.array(truth_settings[0]['index']) - 1
        truth_f_trimmed = truth_f[feature_begin:feature_end, truth_index]

        if predict_data.shape[0] > 0:
            predict_data_trimmed = predict_data[feature_begin:feature_end, :]
            # Prediction Activity analysis to select top prediction class
            # true if active in any frame
            predict_activity = np.any(predict_data_trimmed >= mixdb_out['class_weights_threshold'], axis=0)
            predict_activity_index = np.array([i for i, x in enumerate(predict_activity) if x]) + 1
            logger.info(f'Prediction active in classes based on threshold {mixdb_out["class_weights_threshold"]}:\n'
                        f'{predict_activity_index}')
            predict_mean = np.mean(predict_data_trimmed, axis=0)
            top_active_classes = np.argsort(predict_mean)[::-1] + 1
            logger.info(f'Top 10 active prediction classes by mean:\n{top_active_classes[0:10]}')
            # plot most active prediction
            predict_data_trimmed = predict_data_trimmed[:, top_active_classes[0] - 1]
            logger.info(f'Plotting prediction class {top_active_classes[0]}')

        # Setup plot of target waveform with truth on top
        # TODO
        #  Make a function w/args mixture_trimmed, target_trimmed, truth_f, pr, mixdat, m, truth_index, target_path
        # calculate number of samples per frame for plotting
        num_plot_features = truth_f_trimmed.shape[0]
        num_plot_samples_per_feature = len(target_trimmed) // num_plot_features
        # number of plot samples multiple of frames
        num_plot_samples = num_plot_samples_per_feature * num_plot_features
        # x-axis in sec
        x_seconds = np.arange(num_plot_samples, dtype=np.float32) / sonusai.mixture.SAMPLE_RATE
        # Reshape/extend truth to #samples in waveform
        num_plot_classes = truth_f_trimmed.shape[1]  # Number of plot truth classes
        y_truth_f = np.reshape(np.tile(np.expand_dims(truth_f_trimmed, 1), [1, num_plot_samples_per_feature, 1]),
                               [num_plot_samples, num_plot_classes])

        fig, ax0 = plt.subplots(1, 1, constrained_layout=True, figsize=figsize)
        ax = np.array([ax0], dtype=object)
        plots = []

        # Plot the time-domain waveforms then truth/prediction on second axis
        if mixture_trimmed.shape[0] > 0:
            color = 'mistyrose'
            mix_plot, = ax[0].plot(x_seconds, mixture_trimmed[0:num_plot_samples], color=color, label='MIX')
            ax[0].tick_params(axis='y', labelcolor='red')
            plots.append(mix_plot)

        if target_trimmed.shape[0] > 0:
            color = 'tab:blue'
            tt_plot, = ax[0].plot(x_seconds, target_trimmed[0:num_plot_samples], color=color, label='TAR')
            ax[0].set_ylabel('ampl', color=color)
            ax[0].tick_params(axis='y', labelcolor=color)
            plots.append(tt_plot)

        ax2 = ax[0].twinx()  # instantiate 2nd y-axis that shares the same x-axis

        # Plot first truth
        # TODO Support multi-channel
        if truth_f_trimmed.shape[0] > 0:
            color = 'tab:green'
            label2 = f'truth{truth_index[0] + 1}'
            # we already handled the x-label with ax1
            ax2.set_ylabel(label2, color=color)
            # logger.info(f'Mixture num {m}, truth_index={truth_index}')
            tr_plot, = ax2.plot(x_seconds, y_truth_f[:, 0], color=color, label=label2)
            ax2.set_ylim([-0.05, 1.05])
            ax2.tick_params(axis='y', labelcolor=color)
            plots.append(tr_plot)

        if predict_data.shape[0] > 0:
            color = 'tab:brown'
            label2 = f'prcl{top_active_classes[0]}'
            # we already handled the x-label with ax1
            ax2.set_ylabel(label2, color=color)
            predict_extended = np.reshape(
                np.tile(np.expand_dims(predict_data_trimmed, 1), [1, num_plot_samples_per_feature, 1]),
                [num_plot_samples, num_plot_classes])
            pr_plot, = ax2.plot(x_seconds, predict_extended, color=color, label=label2)
            ax2.set_ylim([-0.05, 1.05])
            ax2.tick_params(axis='y', labelcolor=color)
            plots.append(pr_plot)

        ax[0].set_xlabel('time (s)')  # set only on last/bottom plot

        # Get actual mixture target config parameters
        target_augmentations = mixdb_out['target_augmentations'][mixdb_out['mixtures'][m]['target_augmentation_index']]
        fig.suptitle(
            f'{m + 1} of {num_mixtures}: {target_path}\n'
            f'{target_record["name"]}\n'
            f'Target augmentations: {target_augmentations}\n'
            f'Truth indices: {truth_settings[0]["index"]}\n'
            f'Global Truth Function:Config {truth_settings[0]["function"]} : {truth_settings[0]["config"]}',
            fontsize=10)
        pdf.savefig(fig)
        plt.close(fig)

    # open config file read in as text, then print to last page
    last_page = plt.figure(figsize=figsize)
    last_page.clf()

    option_text = \
        f'Function parameters:\n' \
        f'  root     {root}\n' \
        f'  config   {config_file}\n' \
        f'  target   {target_file}\n' \
        f'  model    {model_file}\n' \
        f'  feature  {feature}\n' \
        f'  sedthr   {thresholds}\n\n' \
        f'----------------------------------------\n\n'
    last_page.text(0.05, 0.95, option_text + yaml.dump(config),
                   transform=last_page.transFigure,
                   family='monospace',
                   size=10,
                   ha="left",
                   va="top")
    pdf.savefig()
    plt.close()
    pdf.close()


def main():
    from docopt import docopt

    import sonusai
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    create_file_handler('tplot.log')
    update_console_handler(verbose)
    initial_log_messages('tplot')

    tplot(root=args['--root'],
          config_file=args['--config'],
          target_file=args['--target'],
          model_file=args['--model'],
          feature=args['--feature'],
          thresholds=args['--sedthr'],
          verbose=verbose)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from dataclasses import dataclass
from os import PathLike
from typing import List
from typing import Union

import numpy as np
from tensorflow.keras.utils import Sequence

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureIDs


class KerasFromH5(Sequence):
    """Generates data for Keras from a directory containing genft data"""

    @dataclass(frozen=True)
    class BatchParams:
        mixid: List[int]
        offset: int
        extra: int

    def __init__(self,
                 mixdb: MixtureDatabase,
                 mixids: MixtureIDs,
                 location: Union[str, bytes, PathLike],
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False):
        """Initialization"""
        from os.path import exists
        from os.path import isdir
        from os.path import join

        import h5py

        from sonusai import SonusAIError
        from sonusai.mixture import convert_mixids_to_list
        from sonusai.mixture import get_feature_stats
        from sonusai.mixture import get_mixid_padded_name

        self.mixdb = mixdb
        self.mixids = convert_mixids_to_list(self.mixdb, mixids)
        self.location = location
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle

        if not isdir(self.location):
            raise SonusAIError(f'Specified location, {self.location}, is not a directory')

        for m in self.mixids:
            mixid_padded_name = get_mixid_padded_name(self.mixdb, m)
            filename = join(self.location, mixid_padded_name + '.h5')
            if not exists(filename):
                raise SonusAIError(f'Could not find {filename}')
            with h5py.File(filename, 'r') as f:
                datasets = ['feature', 'truth_f']
                for dataset in datasets:
                    if dataset not in f:
                        raise SonusAIError(f'Could not find "{dataset}" in {filename}')

        fs = get_feature_stats(feature=self.mixdb.feature,
                               num_classes=self.mixdb.num_classes,
                               frame_size=self.mixdb.frame_size)

        self.stride = fs.stride
        self.num_bands = fs.num_bands
        self.num_classes = self.mixdb.num_classes
        self.mrecords = None
        self.mixture_frame_segments = None
        self.batch_frame_segments = None
        self.total_batches = None

        self._initialize_mixtures()

    def __len__(self) -> int:
        """Denotes the number of batches per epoch"""
        return self.total_batches

    def __getitem__(self, batch_index: int) -> (np.ndarray, np.ndarray):
        """Get one batch of data"""
        from sonusai.utils import reshape_inputs

        batch_params = self.batch_params[batch_index]

        result = []
        for mixid in batch_params.mixid:
            result.append(self._get_ft(mixid))

        feature = np.vstack([result[i][0] for i in range(len(result))])
        truth = np.vstack([result[i][1] for i in range(len(result))])

        if batch_params.extra > 0:
            feature = feature[batch_params.offset:-batch_params.extra]
            truth = truth[batch_params.offset:-batch_params.extra]
        else:
            feature = feature[batch_params.offset:]
            truth = truth[batch_params.offset:]

        feature, truth, _, _, _, _ = reshape_inputs(feature=feature,
                                                    truth=truth,
                                                    batch_size=self.batch_size,
                                                    timesteps=self.timesteps,
                                                    flatten=self.flatten,
                                                    add1ch=self.add1ch)

        return feature, truth

    def on_epoch_end(self) -> None:
        """Modification of dataset between epochs"""
        import random

        if self.shuffle:
            random.shuffle(self.mixids)
            self._initialize_mixtures()

    def _get_ft(self, mixid: int) -> (np.ndarray, np.ndarray):
        from os.path import join

        import h5py

        from sonusai.mixture import get_mixid_padded_name

        mixid_padded_name = get_mixid_padded_name(self.mixdb, mixid)

        with h5py.File(join(self.location, mixid_padded_name + '.h5'), 'r') as f:
            feature = np.array(f['feature'])
            truth_f = np.array(f['truth_f'])

        return feature, truth_f

    def _initialize_mixtures(self) -> None:
        from sonusai.data_generator.utils import get_frames_per_batch_from_mixdb
        from sonusai.mixture import get_mrecords_from_mixids

        self.mrecords = get_mrecords_from_mixids(mixdb=self.mixdb, mixids=self.mixids)
        frames_per_batch, self.total_batches = get_frames_per_batch_from_mixdb(mixdb=self.mixdb,
                                                                               mrecords=self.mrecords,
                                                                               batch_size=self.batch_size,
                                                                               timesteps=self.timesteps)

        # Compute mixid, offset, and extra for dataset
        # offsets and extras are needed because mixtures are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence. And the extra must be maintained in order to preserve the correct data length.
        cumulative_frames = 0
        start_mixture_index = 0
        offset = 0
        self.batch_params = []
        for m in range(len(self.mixids)):
            current_frames = self.mrecords[m].samples // self.mixdb.feature_step_samples
            cumulative_frames += current_frames
            while cumulative_frames >= frames_per_batch:
                extra = cumulative_frames - frames_per_batch
                mixid = self.mixids[start_mixture_index:m + 1]
                self.batch_params.append(self.BatchParams(mixid=mixid, offset=offset, extra=extra))
                if extra == 0:
                    start_mixture_index = m + 1
                    offset = 0
                else:
                    start_mixture_index = m
                    offset = current_frames - extra
                cumulative_frames = extra



"""sonusai keras_train

usage: keras_train [-hv] (-m MODEL) (-d MIXDB) [--tloc TLOC] [-l VMIXDB] [--vloc VLOC]
                    [-e EPOCHS] [-b BATCH] [-t TSTEPS] [-p ESP] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -m MODEL, --model MODEL         Model Python file with build and/or hypermodel functions.
    -d MIXDB, --mixdb MIXDB         Mixture database JSON file.
    --tloc TLOC                     Data location.
    -l VMIXDB, --vmixdb VMIXDB      Validation split (0 - 1) or validation mixdb. [default: 0.2].
    --vloc VLOC                     Validation data location.
    -e EPOCHS, --epochs EPOCHS      Number of epochs to use in training. [default: 8].
    -b BATCH, --batch BATCH         Batch size.
    -t TSTEPS, --tsteps TSTEPS      Timesteps.
    -p ESP, --patience ESP          Early stopping patience.
    -o OUTPUT, --output OUTPUT      Output directory.

Use Keras to train a model defined by a Python definition file and SonusAI genft data.

Inputs:
    MIXDB   A SonusAI mixture database JSON file.
    TLOC    Optional training data directory containing genft data.
    VMIXDB  A validation data split (0 - 1) or a SonusAI mixture database JSON file to use for validation data.
    VLOC    Optional validation data directory containing genft data; only used when VMIXDB is a JSON mixdb.

Validation is created from MIXDB data using a split of VMIXDB if VMIXDB is a number between 0 and 1, or from VMIXDB
if VMIXDB is a SonusAI mixture database JSON file.

Results are written into subdirectory <MODEL>-<TIMESTAMP> unless OUTPUT is specified.

"""
from sonusai import logger


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    model_name = args['--model']
    t_name = args['--mixdb']
    t_loc = args['--tloc']
    v_name = args['--vmixdb']
    v_loc = args['--vloc']
    epochs = int(args['--epochs'])
    batch_size = args['--batch']
    timesteps = args['--tsteps']
    esp = args['--patience']
    output_dir = args['--output']

    from os import makedirs
    from os import walk
    from os.path import basename
    from os.path import join
    from os.path import splitext

    import h5py
    import keras_tuner as kt
    import numpy as np
    import tensorflow as tf
    from tensorflow.keras import backend as kb
    from tensorflow.keras.callbacks import EarlyStopping

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.data_generator import KerasFromH5
    from sonusai.mixture import get_class_count_from_mixids
    from sonusai.mixture import load_mixdb
    from sonusai.mixture import new_mixdb_from_mixids
    from sonusai.utils import check_keras_overrides
    from sonusai.utils import create_ts_name
    from sonusai.utils import import_keras_model
    from sonusai.utils import stratified_shuffle_split_mixid

    v_mixdb = None
    v_split = None
    try:
        v_split = float(v_name)
        if v_split < 0 or v_split > 1:
            logger.exception('--vmixdb must be between 0 and 1')
            raise SystemExit(1)
    except ValueError:
        v_mixdb = load_mixdb(v_name)
        if v_loc is None:
            v_loc = splitext(v_name)[0]

    if t_loc is None:
        t_loc = splitext(t_name)[0]

    model_base = basename(model_name)
    model_root = splitext(model_base)[0]

    if batch_size is not None:
        batch_size = int(batch_size)

    if timesteps is not None:
        timesteps = int(timesteps)

    if output_dir is None:
        output_dir = create_ts_name(model_root)

    makedirs(output_dir, exist_ok=True)
    base_name = join(output_dir, model_root)

    # Setup logging file
    create_file_handler(join(output_dir, 'keras_train.log'))
    update_console_handler(verbose)
    initial_log_messages('keras_train')

    logger.info(f'tensorflow    {tf.__version__}')
    logger.info(f'keras         {tf.keras.__version__}')
    logger.info('')

    t_mixdb = load_mixdb(t_name)
    t_num_mixtures = len(t_mixdb.mixtures)
    logger.info(f'Found {t_num_mixtures} mixtures with {t_mixdb.num_classes} classes from {t_name}')

    # Import model definition file
    logger.info(f'Importing {model_base}')
    model = import_keras_model(model_name)

    # Check overrides
    timesteps = check_keras_overrides(model, t_mixdb.feature, t_mixdb.num_classes, timesteps, batch_size)

    logger.info('Building model')
    try:
        hypermodel = model.MyHyperModel(feature=t_mixdb.feature,
                                        num_classes=t_mixdb.num_classes,
                                        timesteps=timesteps,
                                        batch_size=batch_size)
        built_model = hypermodel.build_model(kt.HyperParameters())
    except Exception as e:
        logger.exception(f'Error: build_model() in {model_base} failed: {e}.')
        raise SystemExit(1)

    kb.clear_session()
    logger.info('')
    built_model.summary(print_fn=logger.info)
    logger.info('')
    logger.info(f'feature       {hypermodel.feature}')
    logger.info(f'num_classes   {hypermodel.num_classes}')
    logger.info(f'batch_size    {hypermodel.batch_size}')
    logger.info(f'timesteps     {hypermodel.timesteps}')
    logger.info(f'flatten       {hypermodel.flatten}')
    logger.info(f'add1ch        {hypermodel.add1ch}')
    logger.info(f'truth_mutex   {hypermodel.truth_mutex}')
    logger.info(f'lossf         {hypermodel.lossf}')
    logger.info(f'input_shape   {hypermodel.input_shape}')
    logger.info(f'optimizer     {built_model.optimizer.get_config()}')
    logger.info('')

    stratify = False
    if v_mixdb is not None:
        # Validation data from v_mixdb
        logger.info(f'Validation data from {v_name}.')
        v_mixid = range(len(v_mixdb.mixtures))
        if stratify:
            logger.info(f'Stratifying training data.')
            t_mixid, _, _, _ = stratified_shuffle_split_mixid(t_mixdb, vsplit=0)
        else:
            t_mixid = range(t_num_mixtures)
    else:
        # Validation data from v_split
        logger.info(f'Validation data from {t_name} using split of {v_split}.')
        v_mixdb = t_mixdb
        v_loc = t_loc
        if stratify:
            logger.info(f'Stratifying training data.')
            t_mixid, v_mixid, _, _ = stratified_shuffle_split_mixid(t_mixdb, vsplit=v_split)
        else:
            t_split_mixtures = int(np.ceil((1 - v_split) * t_num_mixtures))
            a_mixid = list(range(t_num_mixtures))
            rnd_seed = 0
            if rnd_seed is not None:
                # randomize order
                import random
                random.shuffle(a_mixid)
            t_mixid = a_mixid[0:t_split_mixtures]
            v_mixid = a_mixid[t_split_mixtures:]

    # Use SonusAI DataGenerator to create validation feature/truth on the fly
    v_datagen = KerasFromH5(mixdb=v_mixdb,
                            mixids=v_mixid,
                            location=v_loc,
                            batch_size=hypermodel.batch_size,
                            timesteps=hypermodel.timesteps,
                            flatten=hypermodel.flatten,
                            add1ch=hypermodel.add1ch,
                            shuffle=False)

    # Prepare class weighting
    class_count = np.ceil(np.array(get_class_count_from_mixids(t_mixdb, t_mixid)) / t_mixdb.feature_step_samples)
    if t_mixdb.truth_mutex:
        other_weight = 16.0
        logger.info(f'Detected single-label mode (truth_mutex); setting other weight to {other_weight}.')
        class_count[-1] = class_count[-1] / other_weight

    # Use SonusAI DataGenerator to create training feature/truth on the fly
    t_datagen = KerasFromH5(mixdb=t_mixdb,
                            mixids=t_mixid,
                            location=t_loc,
                            batch_size=hypermodel.batch_size,
                            timesteps=hypermodel.timesteps,
                            flatten=hypermodel.flatten,
                            add1ch=hypermodel.add1ch,
                            shuffle=True)

    # TODO: If hypermodel.es exists, then use it; otherwise use default here
    if esp is None:
        es = EarlyStopping(monitor='val_loss',
                           mode='min',
                           verbose=1,
                           patience=8)
    else:
        es = EarlyStopping(monitor='val_loss',
                           mode='min',
                           verbose=1,
                           patience=int(esp))

    ckpt_callback = tf.keras.callbacks.ModelCheckpoint(filepath=base_name + '-ckpt-weights.h5',
                                                       save_weights_only=True,
                                                       monitor='val_loss',
                                                       mode='min',
                                                       save_best_only=True)

    logger.info('')
    logger.info(f'Training with no class weighting and early stopping patience = {es.patience}')
    logger.info(f'  training mixtures    {len(t_mixid)}')
    logger.info(f'  validation mixtures  {len(v_mixid)}')
    logger.info('')

    history = built_model.fit(t_datagen,
                              batch_size=hypermodel.batch_size,
                              epochs=epochs,
                              validation_data=v_datagen,
                              shuffle=False,
                              callbacks=[es, ckpt_callback])

    # Save history into numpy file
    history_name = base_name + '-history'
    np.save(history_name, history.history)
    # Note: Reload with history=np.load(history_name, allow_pickle='TRUE').item()
    logger.info(f'Saved training history to numpy file {history_name}.npy')

    # Find checkpoint file and load weights for prediction and model save
    checkpoint_name = None
    for path, dirs, files in walk(output_dir):
        for f in files:
            if "ckpt" in f:
                checkpoint_name = f

    if checkpoint_name is not None:
        logger.info('Using best checkpoint for prediction and model exports')
        built_model.load_weights(join(output_dir, checkpoint_name))
    else:
        logger.info('Using last epoch for prediction and model exports')

    # save for later model export(s)
    weight_name = base_name + '.h5'
    built_model.save(weight_name)
    with h5py.File(weight_name, 'a') as f:
        f.attrs['sonusai_feature'] = hypermodel.feature
        f.attrs['sonusai_num_classes'] = str(hypermodel.num_classes)
    logger.info(f'Saved trained model to {weight_name}')

    # Compute prediction metrics on validation data using the best checkpoint
    # t_mixid, v_mixid, or entire dataset mixdb
    v_mixdb = new_mixdb_from_mixids(t_mixdb, v_mixid)
    v_predict = built_model.predict(v_datagen, batch_size=hypermodel.batch_size, verbose=1)
    v_predict_name = base_name + '-valpredict.h5'
    with h5py.File(v_predict_name, 'w') as f:
        f.create_dataset('predict', data=v_predict)
        f.attrs['mixdb'] = v_mixdb.to_json()

    logger.info(f'Wrote validation predict data to file {v_predict_name}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



def trim_docstring(docstring: str) -> str:
    """Trim whitespace from docstring"""
    from sys import maxsize

    if not docstring:
        return ''

    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()

    # Determine minimum indentation (first line doesn't count)
    indent = maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))

    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())

    # Strip off leading blank lines:
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    # Return a single string
    return '\n'.join(trimmed)



#!/usr/bin/env python3
"""aawscd_probwrite

usage: aawscd_probwrite [-h] [-m MACHINE] [-f FRAMES] FILE

options:
   -h, --help
   -m MACHINE, --machine MACHINE    IP address of aawscd platform. [default: localhost].
   -f FRAMES, --frames FRAMES       Number of frames to capture. [default: 10].

aawscd_probwrite connects to an Aaware platform running aawscd and writes the sound classification
probability output to an HDF5 file.

"""
from threading import Condition

import numpy as np

CLIENT = 'aawscd_probwrite'
TOPIC = 'aawscd/sc/prob'
DONE = Condition()
FRAMES = 10
FRAME_COUNT = 0
DATA = None
PROGRESS = None


def shutdown(_signum, _frame):
    global DONE
    with DONE:
        DONE.notify()


def unpack_prob_entry(entry: int) -> (int, int, int):
    """Decode the packed probability data: [ 16-bit label | 8-bit zone | 8-bit probability ]."""
    data = np.array(entry, dtype=np.uint32)
    label = np.right_shift(data, 16)
    zone = np.bitwise_and(np.right_shift(data, 8), 0xFF)
    value = np.bitwise_and(entry, 0xFF)
    return int(zone), int(label), int(value)


def parse_prob(payload: list) -> np.ndarray:
    """Parse MQTT probability payload from aawscd.

    The 'aawscd/sc/prob' payload is a list of uint32. Each item contains packed data and is
    unpacked/decoded and inserted into an array.
    """

    # First pass: get the zones and labels to determine the size of the array
    zones = set()
    labels = set()
    for entry in payload:
        zone, label, _ = unpack_prob_entry(entry)
        zones.add(zone)
        labels.add(label)

    # Create the array
    zones = list(zones)
    labels = list(labels)
    prob = np.zeros((len(labels), len(zones)), dtype=np.uint8)

    # Second pass: fill in the array probability values.
    for entry in payload:
        zone, label, value = unpack_prob_entry(entry)
        prob[label, zone] = value

    # Return the array
    return prob


def on_message(_client, _userdata, message):
    import paho.mqtt.client as mqtt
    import yaml

    global TOPIC
    if mqtt.topic_matches_sub(TOPIC, message.topic):
        payload = yaml.safe_load(str(message.payload.decode('utf-8')))
        prob = parse_prob(payload['prob'])

        global DATA
        global FRAMES
        if DATA is None:
            DATA = np.zeros((FRAMES, prob.shape[0], prob.shape[1]), dtype=prob.dtype)

        global FRAME_COUNT
        DATA[FRAME_COUNT] = prob
        FRAME_COUNT += 1

        global PROGRESS
        PROGRESS.update()

        if FRAME_COUNT == FRAMES:
            global DONE
            with DONE:
                DONE.notify()


def main():
    import signal

    import h5py
    import paho.mqtt.client as mqtt
    from docopt import docopt
    from tqdm import tqdm

    args = docopt(__doc__, version='1.0.0', options_first=True)

    machine = args['--machine']

    global FRAMES
    FRAMES = int(args['--frames'])

    file = args['FILE']

    signal.signal(signal.SIGINT, shutdown)
    signal.signal(signal.SIGTERM, shutdown)

    global CLIENT
    client = mqtt.Client(client_id=CLIENT)
    client.on_message = on_message
    client.connect(host=machine)
    client.loop_start()
    global TOPIC
    client.subscribe(topic=TOPIC)

    global PROGRESS
    PROGRESS = tqdm(total=FRAMES, desc=file)

    with DONE:
        DONE.wait()

    PROGRESS.close()

    client.unsubscribe(topic=TOPIC)
    client.loop_stop()
    client.disconnect()

    global DATA
    with h5py.File(file, 'w') as f:
        f.create_dataset(name='prob', data=DATA)

    print(f'Wrote {file}')


if __name__ == '__main__':
    main()



from sonusai.mixture import MRecords
from sonusai.mixture import MixtureDatabase


def get_frames_per_batch_from_mixdb(mixdb: MixtureDatabase,
                                    mrecords: MRecords,
                                    batch_size: int,
                                    timesteps: int) -> (int, int):
    from sonusai import SonusAIError

    frames = sum([sub.samples for sub in mrecords]) // mixdb.feature_step_samples
    frames_per_batch = get_frames_per_batch(batch_size, timesteps)
    total_batches = frames // frames_per_batch
    if total_batches == 0:
        raise SonusAIError(
            f'Error: dataset only contains {frames} frames which is not enough to fill a batch size of '
            f'{frames_per_batch}. Either provide more data or decrease the batch size')

    return frames_per_batch, total_batches


def get_frames_per_batch(batch_size: int, timesteps: int) -> int:
    return batch_size if timesteps == 0 else batch_size * timesteps



from typing import Dict
from typing import List
from typing import Union

import numpy as np

from sonusai.mixture.mixdb import Augmentation
from sonusai.mixture.mixdb import Augmentations
from sonusai.mixture.mixdb import MRecord
from sonusai.mixture.mixdb import evaluate_random_rule


def get_augmentations(rules: Union[List[Dict], Dict]) -> Augmentations:
    """Generate augmentations from list of input rules."""

    from sonusai.utils import dataclass_from_dict

    processed_rules = []
    if not isinstance(rules, list):
        rules = [rules]

    for rule in rules:
        expand_rules(processed_rules, rule)

    processed_rules = randomize_rules(processed_rules)

    augmentations = []
    for processed_rule in processed_rules:
        augmentations.append(dataclass_from_dict(Augmentation, processed_rule))

    return augmentations


def expand_rules(out_rules: List[Dict], in_rule: dict) -> None:
    """Expand rules."""
    from copy import deepcopy
    from numbers import Number

    from sonusai import SonusAIError
    from sonusai.mixture import VALID_AUGMENTATIONS

    for key, value in list(in_rule.items()):
        if value is None:
            del in_rule[key]

    # replace old 'eq' rule with new 'eq1' rule to allow both for backward compatibility
    in_rule = {'eq1' if key == 'eq' else key: value for key, value in in_rule.items()}

    for key in in_rule:
        if key not in VALID_AUGMENTATIONS:
            nice_list = '\n'.join([f'  {item}' for item in VALID_AUGMENTATIONS])
            raise SonusAIError(f'Invalid augmentation: {key}.\nValid augmentations are:\n{nice_list}')

        if key in ['eq1', 'eq2', 'eq3']:
            # eq must be a list of length 3 or a list of length 3 lists
            valid = True
            multiple = False
            if isinstance(in_rule[key], list):
                if any(isinstance(el, list) for el in in_rule[key]):
                    multiple = True
                    for value in in_rule[key]:
                        if not isinstance(value, list) or len(value) != 3:
                            valid = False
                else:
                    if len(in_rule[key]) != 3:
                        valid = False
            else:
                valid = False

            if not valid:
                raise SonusAIError(f'Invalid augmentation value for {key}: {in_rule[key]}')

            if multiple:
                for value in in_rule[key]:
                    expanded_rule = deepcopy(in_rule)
                    expanded_rule[key] = deepcopy(value)
                    expand_rules(out_rules, expanded_rule)
                return

        elif key in ['count', 'mixup']:
            pass

        else:
            if isinstance(in_rule[key], list):
                for value in in_rule[key]:
                    if isinstance(value, list):
                        raise SonusAIError(f'Invalid augmentation value for {key}: {in_rule[key]}')
                    expanded_rule = deepcopy(in_rule)
                    expanded_rule[key] = deepcopy(value)
                    expand_rules(out_rules, expanded_rule)
                return
            elif not isinstance(in_rule[key], Number):
                if not in_rule[key].startswith('rand'):
                    raise SonusAIError(f'Invalid augmentation value for {key}: {in_rule[key]}')

    out_rules.append(in_rule)


def randomize_rules(in_rules: List[Dict]) -> List[Dict]:
    """Randomize rules."""
    out_rules = []
    for in_rule in in_rules:
        if rule_has_rand(in_rule):
            count = 1
            if 'count' in in_rule and in_rule['count'] is not None:
                count = in_rule['count']
                del in_rule['count']
            for i in range(count):
                out_rules.append(generate_random_rule(in_rule))
        else:
            out_rules.append(in_rule)
    return out_rules


def generate_random_rule(in_rule: dict) -> dict:
    """Generate a new rule from a rule that contains 'rand' directives."""
    from copy import deepcopy

    out_rule = deepcopy(in_rule)
    for key in out_rule:
        out_rule[key] = evaluate_random_rule(str(out_rule[key]))

        # convert eq values from strings to numbers
        if key in ['eq1', 'eq2', 'eq3']:
            for n in range(3):
                if isinstance(out_rule[key][n], str):
                    out_rule[key][n] = eval(out_rule[key][n])

    return out_rule


def rule_has_rand(rule: dict) -> bool:
    """Determine if any keys in the given rule contain 'rand'"""
    for key in rule:
        if 'rand' in str(rule[key]):
            return True

    return False


def apply_augmentation(audio_in: np.ndarray,
                       augmentation: Augmentation,
                       length_common_denominator: int = 1) -> np.ndarray:
    """Use sox to apply augmentations to audio data."""

    import sox

    from sonusai import SonusAIError
    from sonusai.mixture import BIT_DEPTH
    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import SAMPLE_RATE

    try:
        # Apply augmentations
        tfm = sox.Transformer()
        tfm.set_input_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)
        tfm.set_output_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)

        # TODO
        #  Always normalize and remove normalize from list of available augmentations
        #  Normalize to globally set level (should this be a global config parameter,
        #  or hard-coded into the script?)
        if augmentation.normalize is not None:
            tfm.norm(db_level=augmentation.normalize)

        if augmentation.gain is not None:
            tfm.gain(gain_db=augmentation.gain, normalize=False)

        if augmentation.pitch is not None:
            tfm.pitch(n_semitones=augmentation.pitch / 100)

        if augmentation.tempo is not None:
            factor = augmentation.tempo
            if abs(factor - 1.0) <= 0.1:
                tfm.stretch(factor=factor)
            else:
                tfm.tempo(factor=factor, audio_type='s')

        if augmentation.eq1 is not None:
            tfm.equalizer(frequency=augmentation.eq1[0], width_q=augmentation.eq1[1],
                          gain_db=augmentation.eq1[2])

        if augmentation.eq2 is not None:
            tfm.equalizer(frequency=augmentation.eq2[0], width_q=augmentation.eq2[1],
                          gain_db=augmentation.eq2[2])

        if augmentation.eq3 is not None:
            tfm.equalizer(frequency=augmentation.eq3[0], width_q=augmentation.eq3[1],
                          gain_db=augmentation.eq3[2])

        if augmentation.lpf is not None:
            tfm.lowpass(frequency=augmentation.lpf)

        # Create output data
        audio_out = tfm.build_array(input_array=audio_in, sample_rate_in=SAMPLE_RATE)

        # make sure length is multiple of length_common_denominator
        audio_out = np.pad(array=audio_out,
                           pad_width=(0, get_pad_length(len(audio_out), length_common_denominator)),
                           mode='constant',
                           constant_values=0)

        return audio_out
    except Exception as e:
        raise SonusAIError(f'Error applying {augmentation}: {e}')


def estimate_audio_length(audio_in: np.ndarray,
                          augmentation: Augmentation,
                          length_common_denominator: int = 1) -> int:
    """Estimate the length of audio after augmentation."""
    length = len(audio_in)

    if augmentation.tempo is not None:
        length = int(length // augmentation.tempo)

    length += get_pad_length(length, length_common_denominator)

    return length


def get_mixups(augmentations: Augmentations) -> List[int]:
    return sorted(list(set([augmentation.mixup for augmentation in augmentations])))


def get_augmentation_indices_for_mixup(augmentations: list, mixup: int) -> list:
    indices = []
    for idx, augmentation in enumerate(augmentations):
        if mixup == augmentation.mixup:
            indices.append(idx)

    return indices


def get_pad_length(length: int, length_common_denominator: int) -> int:
    mod = int(length % length_common_denominator)
    if mod:
        return length_common_denominator - mod
    return 0


def pad_to_samples(audio_in: np.ndarray, samples: int) -> np.ndarray:
    return np.pad(audio_in, (0, samples - len(audio_in)), mode='constant', constant_values=0)


def apply_snr_gain(mrecord: MRecord, targets: List[np.ndarray], noise: np.ndarray) -> (List[np.ndarray], np.ndarray):
    for idx in range(len(targets)):
        targets[idx] = np.int16(np.round(np.single(targets[idx]) * mrecord.target_snr_gain))

    noise = np.int16(np.round(np.single(noise) * mrecord.noise_snr_gain))

    return targets, noise



def log_duration_and_sizes(total_duration: int,
                           num_classes: int,
                           feature_step_samples: int,
                           num_bands: int,
                           stride: int,
                           desc: str) -> None:
    from sonusai import logger
    from sonusai.mixture import BIT_DEPTH
    from sonusai.mixture import FLOAT_BYTES
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms

    total_samples = total_duration * SAMPLE_RATE
    mixture_bytes = total_samples * BIT_DEPTH / 8
    truth_t_bytes = total_samples * num_classes * FLOAT_BYTES
    feature_bytes = total_samples / feature_step_samples * stride * num_bands * FLOAT_BYTES
    truth_f_bytes = total_samples / feature_step_samples * num_classes * FLOAT_BYTES

    logger.info('')
    logger.info(f'{desc} duration:   {seconds_to_hms(seconds=total_duration)}')
    logger.info(f'{desc} sizes:')
    logger.info(f' mixture:             {human_readable_size(mixture_bytes, 1)}')
    logger.info(f' truth_t:             {human_readable_size(truth_t_bytes, 1)}')
    logger.info(f' feature:             {human_readable_size(feature_bytes, 1)}')
    logger.info(f' truth_f:             {human_readable_size(truth_f_bytes, 1)}')



"""sonusai mkwav

usage: mkwav [-hvtn] (-d MIXDB) [-i MIXID] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Mixture database JSON file.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -o OUTPUT, --output OUTPUT      Output directory.
    -t, --target                    Write target file.
    -n, --noise                     Write noise file.

The mkwav command creates WAV files from a SonusAI database.

Inputs:
    MIXDB       A SonusAI mixture database JSON file.
    MIXID       A glob of mixture ID(s) to generate.

Outputs:
    OUTPUT/     A directory containing:
                    <id>_mixture.wav:   mixture
                    <id>_target.wav:    target (optional)
                    <id>_noise.wav:     noise (optional)
                    <id>.txt
                    mkwav.log

"""
from typing import List

import numpy as np

from sonusai import logger
from sonusai.mixture import MixtureDatabase

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = {}


def mkwav(mixdb: MixtureDatabase,
          mixid: int,
          raw_target_audios: List[np.ndarray] = None,
          augmented_noise_audios: List[List[np.ndarray]] = None) -> (np.ndarray, np.ndarray, np.ndarray):
    from sonusai.genmix import genmix

    mixture, _, target, noise, _ = genmix(mixdb=mixdb,
                                          mixids=mixid,
                                          raw_target_audios=raw_target_audios,
                                          augmented_noise_audios=augmented_noise_audios)
    return mixture, target, noise


def _process_mixture(mixid: int) -> None:
    from os.path import exists
    from os.path import join

    import h5py

    from sonusai.mixture import get_mixid_padded_name
    from sonusai.mixture import get_mixture_metadata

    mixdb = MP_DICT['mixdb']
    output_dir = MP_DICT['output_dir']
    raw_target_audios = MP_DICT['raw_target_audios']
    augmented_noise_audios = MP_DICT['augmented_noise_audios']
    write_target = MP_DICT['write_target']
    write_noise = MP_DICT['write_noise']

    mixid_padded_name = get_mixid_padded_name(mixdb, mixid)
    output_base = join(output_dir, mixid_padded_name)

    target = None
    noise = None

    need_data = True
    if exists(output_base + '.h5'):
        with h5py.File(output_base + '.h5', 'r') as f:
            if 'mixture' in f:
                need_data = False
            if write_target and 'target' not in f:
                need_data = True
            if write_noise and 'noise' not in f:
                need_data = True

    if need_data:
        mixture, target, noise = mkwav(mixdb=mixdb,
                                       mixid=mixid,
                                       raw_target_audios=raw_target_audios,
                                       augmented_noise_audios=augmented_noise_audios)
    else:
        with h5py.File(output_base + '.h5', 'r') as f:
            mixture = np.array(f['mixture'])
            if write_target:
                target = np.array(f['target'])
            if write_noise:
                noise = np.array(f['noise'])

    _write_wav(name=output_base + '_mixture.wav', data=mixture)
    if write_target:
        _write_wav(name=output_base + '_target.wav', data=target)
    if write_noise:
        _write_wav(name=output_base + '_noise.wav', data=noise)

    with open(file=join(output_dir, mixid_padded_name + '.txt'), mode='w') as f:
        f.write(get_mixture_metadata(mixdb, mixid))


def _write_wav(name: str, data: np.ndarray) -> None:
    import wave

    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import SAMPLE_BYTES
    from sonusai.mixture import SAMPLE_RATE

    with wave.open(name, mode='w') as f:
        f.setnchannels(CHANNEL_COUNT)
        f.setsampwidth(SAMPLE_BYTES)
        f.setframerate(SAMPLE_RATE)
        f.writeframesraw(data)


def main():
    import time
    from os import makedirs
    from os import remove
    from os.path import exists
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from docopt import docopt
    from tqdm import tqdm

    import sonusai
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import get_augmented_noise_audios
    from sonusai.mixture import get_raw_audios
    from sonusai.mixture import check_audio_files_exist
    from sonusai.mixture import convert_mixids_to_list
    from sonusai.mixture import load_mixdb
    from sonusai.utils import p_tqdm_map
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixdb_name = args['--mixdb']
    mixid = args['--mixid']
    output_dir = args['--output']
    write_target = args['--target']
    write_noise = args['--noise']

    if not output_dir:
        output_dir = splitext(mixdb_name)[0]

    if exists(output_dir) and not isdir(output_dir):
        remove(output_dir)

    makedirs(output_dir, exist_ok=True)

    start_time = time.monotonic()

    create_file_handler(join(output_dir, 'mkwav.log'))
    update_console_handler(verbose)
    initial_log_messages('mkwav')

    logger.info(f'\nLoad mixture database from {mixdb_name}')
    mixdb = load_mixdb(name=mixdb_name)
    mixid = convert_mixids_to_list(mixdb, mixid)

    total_samples = sum([sub.samples for sub in [mixdb.mixtures[m] for m in mixid]])
    duration = total_samples / sonusai.mixture.SAMPLE_RATE

    logger.info('')
    logger.info(f'Found {len(mixid):,} mixtures to process')
    logger.info(f'{total_samples:,} samples')

    check_audio_files_exist(mixdb)

    MP_DICT['mixdb'] = mixdb
    MP_DICT['output_dir'] = output_dir
    MP_DICT['raw_target_audios'] = get_raw_audios(mixdb)
    MP_DICT['augmented_noise_audios'] = get_augmented_noise_audios(mixdb)
    MP_DICT['write_target'] = write_target
    MP_DICT['write_noise'] = write_noise

    progress = tqdm(total=len(mixid), desc='mkwav')
    p_tqdm_map(_process_mixture, mixid, progress=progress)
    progress.close()

    logger.info(f'Wrote {len(mixid)} mixtures to {output_dir}')
    logger.info('')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'mixture:  {human_readable_size(total_samples * 2, 1)}')
    if write_target:
        logger.info(f'target:   {human_readable_size(total_samples * 2, 1)}')
    if write_noise:
        logger.info(f'noise:    {human_readable_size(total_samples * 2, 1)}')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



"""sonusai keras_onnx

usage: keras_onnx [-hvr] (-m MODEL) (-w WEIGHTS) [-b BATCH] [-t TSTEPS] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -m MODEL, --model MODEL         Python model file.
    -w WEIGHTS, --weights WEIGHTS   Keras model weights file.
    -b BATCH, --batch BATCH         Batch size.
    -t TSTEPS, --tsteps TSTEPS      Timesteps.
    -o OUTPUT, --output OUTPUT      Output directory.

Convert a trained Keras model to ONNX.

Inputs:
    MODEL       A SonusAI Python model file with build and/or hypermodel functions.
    WEIGHTS     A Keras model weights file (or model file with weights).

Outputs:
    OUTPUT/     A directory containing:
                    <MODEL>.onnx        Model file with batch_size and timesteps equal to provided parameters
                    <MODEL>-b1.onnx     Model file with batch_size=1 and if the timesteps dimension exists it
                                        is set to 1 (useful for real-time inference applications)
                    keras_onnx.log

Results are written into subdirectory <MODEL>-<TIMESTAMP> unless OUTPUT is specified.

"""
from sonusai import logger


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    model_name = args['--model']
    weight_name = args['--weights']
    batch_size = args['--batch']
    timesteps = args['--tsteps']
    output_dir = args['--output']

    from os import makedirs
    from os.path import basename
    from os.path import join
    from os.path import splitext

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.utils import create_ts_name
    from sonusai.utils import keras_onnx

    model_tail = basename(model_name)
    model_root = splitext(model_tail)[0]

    if batch_size is not None:
        batch_size = int(batch_size)

    if timesteps is not None:
        timesteps = int(timesteps)

    if output_dir is None:
        output_dir = create_ts_name(model_root)

    makedirs(output_dir, exist_ok=True)

    # Setup logging file
    create_file_handler(join(output_dir, 'keras_onnx.log'))
    update_console_handler(verbose)
    initial_log_messages('keras_onnx')

    keras_onnx(model_name, weight_name, timesteps, batch_size, output_dir)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



"""sonusai plot

usage: plot [-hve] [-m MODEL] [-l CSV] [-i MIXID] [-o OUTPUT] INPUT

options:
   -h, --help
   -v, --verbose                Be verbose.
   -m MODEL, --model MODEL      Trained model ONNX file.
   -i MIXID, --mixid MIXID      Mixture to plot if input is HDF5.
   -l CSV, --labels CSV         Optional CSV file of class labels (from SonusAI gentcst).
   -o OUTPUT, --output OUTPUT   Optional output HDF5 file for prediction.
   -e, --energy                 Use energy plots.

Plot SonusAI audio, feature, truth, and prediction data. INPUT must be one of three file types:

    * WAV
      Using the given model, generate feature data and run prediction. A model file must be
      provided. The MIXID is ignored. If --energy is specified, plot predict data as energy.

    * JSON
      Using the given mixture database, generate feature and truth data. Run prediction if
      a model is given. The MIXID is required. (--energy is ignored.)

    * HDF5
      Using the given mixture database, generate feature and truth data. Run prediction if
      a model is given. The MIXID is required. (--energy is ignored.)

Prediction data will be written to OUTPUT if a model file is given and OUTPUT is specified.

There will be one plot per active truth index. In addition, the top 5 prediction classes are determined and
plotted if needed (i.e., if they were not already included in the truth plots). For plots generated using a
mixture database, then the target will also be displayed. If mixup is active, then each target involved will
be added to the corresponding truth plot.

Inputs:
    MODEL   A SonusAI trained ONNX model file. If a model file is given, prediction data will be
            generated.
    INPUT   A WAV file, or
            a JSON file containing a SonusAI mixture database, or
            an HDF5 file containing:
                attribute:  mixdb (required)

Outputs:
    {INPUT}-plot.pdf or {INPUT}-mix{MIXID}-plot.pdf
    plot.log
    OUTPUT (if MODEL and OUTPUT are both specified)

"""
from typing import List
from typing import Union

import matplotlib.pyplot as plt
import numpy as np

from sonusai import logger


def spec_plot(mixture: np.ndarray,
              feature: np.ndarray,
              predict: Union[np.ndarray, None] = None,
              labels: Union[List[str], None] = None,
              title: str = '') -> plt.figure:
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import int16_to_float

    num_plots = 4 if predict is not None else 2
    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[0].plot(x_axis, int16_to_float(mixture), label='Mixture')
    ax[0].set_xlim(x_axis[0], x_axis[-1])
    ax[0].set_ylim([-1.025, 1.025])
    ax[0].set_title('Waveform')

    # Plot the spectrogram
    ax[1].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
    ax[1].set_title('Feature')

    if predict is not None:
        # Plot and label the model output scores for the top-scoring classes.
        mean_predict = np.mean(predict, axis=0)
        num_classes = predict.shape[-1]
        top_n = min(10, num_classes)
        top_class_indices = np.argsort(mean_predict)[::-1][:top_n]
        ax[2].imshow(np.transpose(predict[:, top_class_indices]), aspect='auto', interpolation='nearest', cmap='gray_r')
        y_ticks = range(0, top_n)
        ax[2].set_yticks(y_ticks, [labels[top_class_indices[x]] for x in y_ticks])
        ax[2].set_ylim(-0.5 + np.array([top_n, 0]))
        ax[2].set_title('Class Scores')

        # Plot the probabilities
        ax[3].plot(predict[:, top_class_indices])
        ax[3].legend(np.array(labels)[top_class_indices], loc='best')
        ax[3].set_title('Class Probabilities')

    fig.suptitle(title)

    return fig


def spec_energy_plot(mixture: np.ndarray,
                     feature: np.ndarray,
                     truth_f: Union[np.ndarray, None] = None,
                     predict: Union[np.ndarray, None] = None) -> plt.figure:
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import int16_to_float

    num_plots = 2
    if truth_f is not None:
        num_plots += 1
    if predict is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    p = 0
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[p].plot(x_axis, int16_to_float(mixture), label='Mixture')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    ax[p].set_ylim([-1.025, 1.025])
    ax[p].set_title('Waveform')

    # Plot the spectrogram
    p += 1
    ax[p].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Feature')

    if truth_f is not None:
        p += 1
        ax[p].imshow(np.transpose(truth_f), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth')

    if predict is not None:
        p += 1
        ax[p].imshow(np.transpose(predict), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Predict')

    return fig


def class_plot(mixture: np.ndarray,
               target: Union[np.ndarray, None] = None,
               truth_f: Union[np.ndarray, None] = None,
               predict: Union[np.ndarray, None] = None,
               label: str = '') -> plt.figure:
    """
    Plot mixture waveform with optional prediction and/or truth together in a single plot.
    The target waveform can optionally be provided, and prediction and truth can have multiple classes.

    Inputs:
      mixture       required, numpy array (samples, 1)
      target        optional, list of numpy arrays (samples, 1)
      truth_f       optional, numpy array (frames, 1)
      predict       optional, numpy array (frames, 1)
      label         optional, label name to use when plotting

    """
    from sonusai import SonusAIError
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import int16_to_float

    if mixture.ndim != 1:
        raise SonusAIError('Too many dimensions in mixture')

    if target is not None and target.ndim != 1:
        raise SonusAIError('Too many dimensions in target')

    # Set default to 1 frame when there is no truth or predict data
    frames = 1
    if truth_f is not None and predict is not None:
        if truth_f.ndim != 1:
            raise SonusAIError('Too many dimensions in truth_f')
        t_frames = len(truth_f)

        if predict.ndim != 1:
            raise SonusAIError('Too many dimensions in predict')
        p_frames = len(predict)

        frames = min(t_frames, p_frames)
    elif truth_f is not None:
        if truth_f.ndim != 1:
            raise SonusAIError('Too many dimensions in truth_f')
        frames = len(truth_f)
    elif predict is not None:
        if predict.ndim != 1:
            raise SonusAIError('Too many dimensions in predict')
        frames = len(predict)

    samples = (len(mixture) // frames) * frames

    # x-axis in sec
    x_axis = np.arange(samples, dtype=np.float32) / SAMPLE_RATE

    fig, ax = plt.subplots(1, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the time-domain waveforms then truth/prediction on second axis
    ax.plot(x_axis, int16_to_float(mixture[0:samples]), color='mistyrose', label='Mixture')
    color = 'red'
    ax.set_xlim(x_axis[0], x_axis[-1])
    ax.set_ylim([-1.025, 1.025])
    ax.set_ylabel(f'Amplitude', color=color)
    ax.tick_params(axis='y', labelcolor=color)

    # Plot target time-domain waveform
    if target is not None:
        ax.plot(x_axis, int16_to_float(target[0:samples]), color='blue', label='Target')

    # instantiate 2nd y-axis that shares the same x-axis
    if truth_f is not None or predict is not None:
        y_label = 'Truth/Predict'
        if truth_f is None:
            y_label = 'Predict'
        if predict is None:
            y_label = 'Truth'

        ax2 = ax.twinx()

        color = 'black'
        ax2.set_xlim(x_axis[0], x_axis[-1])
        ax2.set_ylim([-0.025, 1.025])
        ax2.set_ylabel(y_label, color=color)
        ax2.tick_params(axis='y', labelcolor=color)

        if truth_f is not None:
            ax2.plot(x_axis, expand_frames_to_samples(truth_f, samples), color='green', label='Truth')

        if predict is not None:
            ax2.plot(x_axis, expand_frames_to_samples(predict, samples), color='brown', label='Predict')

    # set only on last/bottom plot
    ax.set_xlabel('time (s)')

    fig.suptitle(label)

    return fig


def expand_frames_to_samples(x: np.ndarray, samples: int) -> np.ndarray:
    samples_per_frame = samples // len(x)
    return np.reshape(np.tile(np.expand_dims(x, 1), [1, samples_per_frame]), samples)


def main():
    from os.path import basename
    from os.path import exists
    from os.path import splitext

    import h5py
    from docopt import docopt
    from matplotlib.backends.backend_pdf import PdfPages
    from pyaaware import Predict

    import sonusai
    from sonusai import SonusAIError
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import get_feature_from_audio
    from sonusai.mixture import get_feature_stats
    from sonusai.mixture import get_mixture_data
    from sonusai.mixture import get_truth_indices_for_mixid
    from sonusai.mixture import load_mixdb
    from sonusai.mixture import read_audio
    from sonusai.utils import get_label_names
    from sonusai.utils import print_mixture_details
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    model_name = args['--model']
    output_name = args['--output']
    labels_name = args['--labels']
    mixid = args['--mixid']
    input_name = args['INPUT']
    energy = args['--energy']

    if mixid is not None:
        mixid = int(mixid)

    create_file_handler('plot.log')
    update_console_handler(verbose)
    initial_log_messages('plot')

    if not exists(input_name):
        raise SonusAIError(f'{input_name} does not exist')

    logger.info('')
    logger.info(f'Input:  {input_name}')
    if model_name is not None:
        logger.info(f'Model:  {model_name}')
    if output_name is not None:
        logger.info(f'Output: {output_name}')
    logger.info('')

    ext = splitext(input_name)[1]

    model = None
    mixdb = None
    target = None
    truth_f = None
    t_indices = None

    if model_name is not None:
        model = Predict(model_name)

    if ext == '.wav':
        if model_name is None:
            raise SonusAIError('Must specify MODEL when input is WAV')

        mixture = read_audio(input_name)
        feature = get_feature_from_audio(audio=mixture, feature=model.feature)
        fs = get_feature_stats(feature=model.feature, num_classes=model.output_shape[-1])
        step = fs.step

        logger.debug(f'Audio samples      {len(mixture)}')
        logger.debug(f'Feature shape      {feature.shape}')

    elif ext in ['.json', '.h5']:
        if mixid is None:
            raise SonusAIError('Must specify MIXID when input is JSON or H5')

        mixdb = load_mixdb(input_name)

        print_mixture_details(mixdb=mixdb, mixid=mixid, desc_len=24, print_fn=logger.info, all_class_counts=False)

        logger.info(f'Generating data for mixture {mixid}')
        mixture, target, _, feature, truth_f = get_mixture_data(mixdb=mixdb, mixid=mixid)
        fs = get_feature_stats(feature=mixdb.feature, num_classes=mixdb.num_classes)
        step = fs.step
        t_indices = [x - 1 for x in get_truth_indices_for_mixid(mixdb=mixdb, mixid=mixid)]

        targets = [mixdb.targets[tfi] for tfi in mixdb.mixtures[mixid].target_file_index]
        truth_functions = list(set([sub2.function for sub1 in targets for sub2 in sub1.truth_settings]))
        energy = 'energy_f' in truth_functions or 'snr_f' in truth_functions

        logger.debug(f'Audio samples      {len(mixture)}')
        logger.debug(f'Targets:')
        mixture_record = mixdb.mixtures[mixid]
        for n in range(len(mixture_record.target_file_index)):
            name = mixdb.targets[mixture_record.target_file_index[n]].name
            duration = mixdb.targets[mixture_record.target_file_index[n]].duration
            augmentation = mixdb.target_augmentations[mixture_record.target_augmentation_index[n]]
            logger.debug(f'  Name             {name}')
            logger.debug(f'  Duration         {duration}')
            logger.debug(f'  Augmentation     {augmentation}')

        logger.debug(f'Feature shape      {feature.shape}')
        logger.debug(f'Truth shape        {truth_f.shape}')

    else:
        raise SonusAIError(f'Unknown file type for {input_name}')

    predict = None
    labels = None
    indices = []
    if model is not None:
        logger.debug('')
        logger.info(f'Running prediction on mixture {mixid}')
        logger.debug(f'Model feature name {model.feature}')
        logger.debug(f'Model input shape  {model.input_shape}')
        logger.debug(f'Model output shape {model.output_shape}')

        if feature.shape[0] < model.input_shape[0]:
            raise SonusAIError(f'Mixture {mixid} contains {feature.shape[0]} '
                               f'frames of data which is not enough to run prediction; '
                               f'at least {model.input_shape[0]} frames are needed for this model.\n'
                               f'Consider using a model with a smaller batch size or a mixture with more data.')

        predict = model.execute(feature)

        labels = get_label_names(num_labels=predict.shape[1], file=labels_name)

        # Report the highest-scoring classes and their scores.
        p_max = np.max(predict, axis=0)
        p_indices = np.argsort(p_max)[::-1][:5]
        p_max_len = max([len(labels[i]) for i in p_indices])

        logger.info('Top 5 active prediction classes by max:')
        for p_index in p_indices:
            logger.info(f'  {labels[p_index]:{p_max_len}s} {p_max[p_index]:.3f}')
        logger.info('')

        indices = list(p_indices)

    # Add truth indices for target (if needed)
    for t_index in t_indices:
        if t_index not in indices:
            indices.append(t_index)

    base_name = basename(splitext(input_name)[0])
    if mixdb is not None:
        title = f'{input_name} Mixture {mixid}'
        pdf_name = f'{base_name}-mix{mixid}-plot.pdf'
    else:
        title = f'{input_name}'
        pdf_name = f'{base_name}-plot.pdf'

    # Original size (frames, stride, num_bands)
    # Decimate in the stride dimension
    # Reshape to get frames*decimated_stride, num_bands
    if feature.ndim != 3:
        raise SonusAIError(f'feature does not have 3 dimensions: frames, stride, num_bands')
    spectrogram = feature[:, -step:, :]
    spectrogram = np.reshape(spectrogram, (spectrogram.shape[0] * spectrogram.shape[1], spectrogram.shape[2]))

    with PdfPages(pdf_name) as pdf:
        pdf.savefig(spec_plot(mixture=mixture,
                              feature=spectrogram,
                              predict=predict,
                              labels=labels,
                              title=title))
        for index in indices:
            if energy:
                t_tmp = None
                if truth_f is not None:
                    t_tmp = 10 * np.log10(truth_f + np.finfo(np.float32).eps)

                p_tmp = None
                if predict is not None:
                    p_tmp = 10 * np.log10(predict + np.finfo(np.float32).eps)

                pdf.savefig(spec_energy_plot(mixture=mixture,
                                             feature=spectrogram,
                                             truth_f=t_tmp,
                                             predict=p_tmp))
            else:
                p_tmp = None
                if predict is not None:
                    p_tmp = predict[:, index]

                l_tmp = None
                if labels is not None:
                    l_tmp = labels[index]

                pdf.savefig(class_plot(mixture=mixture,
                                       target=target[index],
                                       truth_f=truth_f[:, index],
                                       predict=p_tmp,
                                       label=l_tmp))
        logger.info(f'Wrote {pdf_name}')

    if output_name:
        with h5py.File(output_name, 'w') as f:
            f.create_dataset(name='predict', data=predict)
            logger.info(f'Wrote {output_name}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from typing import List
from typing import Tuple
from typing import Union

import numpy as np

from sonusai.mixture.mixdb import MRecord
from sonusai.mixture.mixdb import MixtureDatabase

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = {}


def get_augmented_noise_audios(mixdb: MixtureDatabase, show_progress: bool = False) -> List[List[np.ndarray]]:
    """Get a list of lists of augmented noise audio data."""
    from tqdm import tqdm

    from sonusai.utils import p_tqdm_map

    raw_noise_audios = get_raw_audios(mixdb=mixdb, show_progress=show_progress, target=False)

    names = [mixdb.noises[i].name for i in range(len(mixdb.noises))]
    progress = tqdm(total=len(names), desc='Augment noise audio', disable=not show_progress)

    MP_DICT['mixdb'] = mixdb

    return p_tqdm_map(augment_audio, raw_noise_audios, progress=progress)


def get_raw_audios(mixdb: MixtureDatabase, show_progress: bool = False, target: bool = True) -> List[np.ndarray]:
    """Get a list of raw target audio data."""
    from tqdm import tqdm

    from sonusai.utils import p_tqdm_map

    if target:
        names = [mixdb.targets[i].name for i in range(len(mixdb.targets))]
        progress = tqdm(total=len(names), desc='Read target audio', disable=not show_progress)
    else:
        names = [mixdb.noises[i].name for i in range(len(mixdb.noises))]
        progress = tqdm(total=len(names), desc='Read noise audio', disable=not show_progress)

    raw_audios = p_tqdm_map(read_audio, names, progress=progress)

    return raw_audios


def check_audio_files_exist(mixdb: MixtureDatabase) -> None:
    """Walk through all the noise and target audio files in a mixture database ensuring that they exist."""
    from os.path import exists

    from sonusai import SonusAIError
    from sonusai.mixture import tokenized_expandvars

    for file_index in range(len(mixdb.noises)):
        file_name, _ = tokenized_expandvars(mixdb.noises[file_index].name)
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')

    for file_index in range(len(mixdb.targets)):
        file_name, _ = tokenized_expandvars(mixdb.targets[file_index].name)
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')


def read_raw_target_audio(mixdb: MixtureDatabase, show_progress: bool = False) -> List[np.ndarray]:
    """Read in all audio data beforehand to avoid reading it multiple times in a loop."""
    from tqdm import tqdm

    from sonusai.utils import p_tqdm_map

    names = [target.name for target in mixdb.targets]
    progress = tqdm(total=len(names), desc='Read target audio', disable=not show_progress)
    raw_target_audio = p_tqdm_map(read_audio, names, progress=progress)
    progress.close()

    return raw_target_audio


def get_target_noise_audio(mixdb: MixtureDatabase,
                           mrecord: MRecord,
                           raw_target_audios: List[np.ndarray] = None,
                           augmented_noise_audios: List[List[np.ndarray]] = None) -> (List[np.ndarray], np.ndarray):
    """Apply augmentations and return augmented target and noise data."""
    from sonusai import SonusAIError
    from sonusai.mixture import apply_augmentation
    from sonusai.mixture import apply_snr_gain
    from sonusai.mixture import pad_to_samples

    if raw_target_audios is None:
        raw_target_audios = get_raw_audios(mixdb)
    if augmented_noise_audios is None:
        augmented_noise_audios = get_augmented_noise_audios(mixdb)

    target_file_index = mrecord.target_file_index
    target_augmentation_index = mrecord.target_augmentation_index
    if len(target_file_index) != len(target_augmentation_index):
        raise SonusAIError('target_file_index and target_augmentation_index are not the same length')

    if mrecord.samples % mixdb.frame_size != 0:
        raise SonusAIError(f'Number of samples in mixture is not a multiple of {mixdb.frame_size}')

    targets = []
    for idx in range(len(target_file_index)):
        target_augmentation = mixdb.target_augmentations[target_augmentation_index[idx]]

        target = apply_augmentation(audio_in=raw_target_audios[target_file_index[idx]],
                                    augmentation=target_augmentation,
                                    length_common_denominator=mixdb.feature_step_samples)

        target = pad_to_samples(audio_in=target, samples=mrecord.samples)
        targets.append(target)

    noise = augmented_noise_audios[mrecord.noise_file_index][mrecord.noise_augmentation_index]
    noise, _ = get_next_noise(offset_in=mrecord.noise_offset,
                              length=mrecord.samples,
                              audio_in=noise)

    return apply_snr_gain(mrecord=mrecord, targets=targets, noise=noise)


def get_audio_and_truth_t(mixdb: MixtureDatabase,
                          mrecord: MRecord,
                          raw_target_audios: List[np.ndarray] = None,
                          augmented_noise_audios: List[List[np.ndarray]] = None,
                          compute_truth: bool = True,
                          compute_segsnr: bool = False,
                          frame_based_segsnr: bool = False) -> (np.ndarray,
                                                                np.ndarray,
                                                                np.ndarray,
                                                                np.ndarray,
                                                                np.ndarray):
    from sonusai.mixture import generate_segsnr
    from sonusai.mixture import generate_truth

    targets, noise = get_target_noise_audio(mixdb=mixdb,
                                            mrecord=mrecord,
                                            raw_target_audios=raw_target_audios,
                                            augmented_noise_audios=augmented_noise_audios)

    truth_t = generate_truth(mixdb=mixdb,
                             mrecord=mrecord,
                             target_audios=targets,
                             noise_audio=noise,
                             compute=compute_truth)

    target = sum(targets)

    segsnr = generate_segsnr(mixdb=mixdb,
                             mrecord=mrecord,
                             target_audio=target,
                             noise_audio=noise,
                             compute=compute_segsnr,
                             frame_based=frame_based_segsnr)

    mixture = target + noise
    return mixture, truth_t, target, noise, segsnr


def get_next_noise(offset_in: int, length: int, audio_in: np.ndarray) -> (np.ndarray, int):
    audio_out = np.take(audio_in, range(offset_in, offset_in + length), mode='wrap')
    offset_out = (offset_in + length) % len(audio_in)
    return audio_out, offset_out


def read_audio(name: str) -> np.ndarray:
    import sox

    from sonusai import SonusAIError
    from sonusai import logger
    from sonusai.mixture import tokenized_expandvars
    from sonusai.mixture import BIT_DEPTH
    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import SAMPLE_RATE

    expanded_name, _ = tokenized_expandvars(name)

    try:
        # Read in and convert to desired format
        inp = sox.Transformer()
        inp.set_output_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)
        return inp.build_array(input_filepath=expanded_name,
                               sample_rate_in=int(sox.file_info.sample_rate(expanded_name)))

    except Exception as e:
        if name != expanded_name:
            logger.error(f'Error reading {name} (expanded: {expanded_name}): {e}')
        else:
            raise SonusAIError(f'Error reading {name}: {e}')


def augment_audio(audio: np.ndarray, target: bool = False) -> List[np.ndarray]:
    from sonusai.mixture import apply_augmentation

    if target:
        augmentations = MP_DICT['mixdb'].target_augmentations
    else:
        augmentations = MP_DICT['mixdb'].noise_augmentations

    augmented_audios = []
    for augmentation in augmentations:
        augmented_audios.append(apply_augmentation(audio_in=audio, augmentation=augmentation))

    return augmented_audios


def get_mixture_data(mixdb: MixtureDatabase,
                     mrecord: MRecord) -> (np.ndarray,
                                           List[Union[np.ndarray, None]],
                                           np.ndarray,
                                           np.ndarray,
                                           np.ndarray):
    """Get mixture data assuming nothing has been loaded into memory already."""

    from sonusai.mixture import apply_augmentation
    from sonusai.mixture import generate_truth
    from sonusai.mixture import get_feature_and_truth_f
    from sonusai.mixture import get_truth_indices_for_target
    from sonusai.mixture import pad_to_samples

    target_audios = []
    target_truth_indices = []
    for idx in range(len(mrecord.target_file_index)):
        target_name = mixdb.targets[mrecord.target_file_index[idx]].name
        target_augmentation = mixdb.target_augmentations[mrecord.target_augmentation_index[idx]]

        target_audio = apply_augmentation(audio_in=read_audio(target_name),
                                          augmentation=target_augmentation,
                                          length_common_denominator=mixdb.feature_step_samples)

        target_audio = np.int16(np.round(np.single(target_audio) * mrecord.target_snr_gain))
        target_audio = pad_to_samples(audio_in=target_audio, samples=mrecord.samples)
        target_audios.append(target_audio)
        target_truth_indices.append(get_truth_indices_for_target(mixdb.targets[mrecord.target_file_index[idx]]))

    augmented_noise_audios = get_augmented_noise_audios(mixdb=mixdb, show_progress=False)
    raw_noise_audio = augmented_noise_audios[mrecord.noise_file_index][mrecord.noise_augmentation_index]
    noise_audio, _ = get_next_noise(offset_in=mrecord.noise_offset,
                                    length=mrecord.samples,
                                    audio_in=raw_noise_audio)

    noise_audio = np.int16(np.round(np.single(noise_audio) * mrecord.noise_snr_gain))

    truth_t = generate_truth(mixdb=mixdb,
                             mrecord=mrecord,
                             target_audios=target_audios,
                             noise_audio=noise_audio,
                             compute=True)

    mixture_audio = sum(target_audios) + noise_audio

    # Transform target_audios into a list num_classes long such that each entry is the target data per class
    class_audio = []
    for n in range(mixdb.num_classes):
        class_audio.append(None)
        for idx in range(len(target_audios)):
            if n + 1 in target_truth_indices[idx]:
                if class_audio[n] is None:
                    class_audio[n] = target_audios[idx]
                else:
                    class_audio[n] += target_audios[idx]

    feature, truth_f, _ = get_feature_and_truth_f(mixdb=mixdb,
                                                  mrecord=mrecord,
                                                  mixture=mixture_audio,
                                                  truth_t=truth_t)

    return mixture_audio, class_audio, noise_audio, feature, truth_f


def extract_results(results: List[Tuple[np.ndarray, ...]]) -> Tuple[np.ndarray, ...]:
    result = ()
    for j in range(len(results[0])):
        if results[0][j] is None:
            result += (None,)
        else:
            result += (np.concatenate([results[i][j] for i in range(len(results))]),)

    return result



import numpy as np


def reshape_inputs(feature: np.ndarray,
                   truth: np.ndarray = None,
                   batch_size: int = None,
                   timesteps: int = 0,
                   flatten: bool = False,
                   add1ch: bool = False) -> (np.ndarray, np.ndarray, tuple, int, np.ndarray, str):
    """Check SonusAI feature and truth data and reshape feature of size frames x strides x bands into
    one of several options:

    If timesteps > 0: (i.e. for recurrent NNs):
      no-flatten, no-channel:   sequences x timesteps x strides       x bands     (4-dim)
      flatten, no-channel:      sequences x timesteps x strides*bands             (3-dim)
      no-flatten, add-1channel: sequences x timesteps x strides       x bands x 1 (5-dim)
      flatten, add-1channel:    sequences x timesteps x strides*bands         x 1 (4-dim)

    If batch_size is None, then do not reshape; just calculate new input shape and return.

    If timesteps == 0, then do not add timesteps dimension.

    The number of samples is trimmed to be a multiple of batch_size (Keras requirement) for
    both feature and truth.
    Channel is added to last/outer dimension for channel_last support in Keras/TF.

    Returns:
      feature       reshaped feature
      truth         reshaped truth
      in_shape      input shape for model (timesteps x feature)
      num_classes   number of classes in truth = output length of nn model
      class_weights weights of each class in truth, per sklearn compute_weights()
      msg           string with report with info on data and operations done
    """
    from sonusai import SonusAIError
    from sonusai.metrics import calculate_class_weights_from_truth

    frames, strides, bands = feature.shape
    if truth is not None:
        truth_frames, num_classes = truth.shape
        # Double-check correctness of inputs
        if frames != truth_frames:
            raise SonusAIError('Frames in feature and truth do not match')
    else:
        num_classes = None

    msg = f'Feature shape={feature.shape}\n'
    if truth is not None:
        msg += f'truth shape={truth.shape}\n'
    msg += f'Reshape request: timesteps={timesteps}, batch_size={batch_size}, flatten={flatten}, add1ch={add1ch}\n'

    # Compute class weights by hand as sklearn does not handle non-existent classes
    if truth is not None:
        class_weights = calculate_class_weights_from_truth(truth)
    else:
        class_weights = None

    # Calculate new input shape only and return
    if batch_size is None:
        if flatten:
            in_shape = [strides * bands]
        else:
            in_shape = [strides, bands]

        if timesteps > 0:
            in_shape = np.concatenate(([timesteps], in_shape[0:]), axis=0)

        if add1ch:
            in_shape = np.concatenate((in_shape[0:], [1]), axis=0)

        return feature, truth, in_shape, num_classes, class_weights, msg  # quick

    if flatten:
        msg += f'Flattening {strides}x{bands} feature to {strides * bands}\n'
        feature = np.reshape(feature, (frames, strides * bands))

    # Reshape for Keras/TF recurrent models that require timesteps/sequence length dimension
    if timesteps > 0:
        sequences = frames // timesteps

        # Remove frames if remainder exists (not fitting into a multiple of new number of sequences)
        frames_rem = frames % timesteps
        batch_rem = (frames // timesteps) % batch_size
        bf_rem = batch_rem * timesteps
        sequences = sequences - batch_rem
        fr2drop = frames_rem + bf_rem
        if fr2drop:
            msg += f'Dropping {fr2drop} frames for new number of sequences to fit in multiple of batch_size\n'
            if feature.ndim == 2:
                feature = feature[0:-fr2drop, ]  # Flattened input
            elif feature.ndim == 3:
                feature = feature[0:-fr2drop, ]  # Unflattened input

            if truth is not None:
                truth = truth[0:-fr2drop, ]

        # Reshape
        msg += f'Reshape for timesteps = {timesteps}, new number of sequences (batches) = {sequences}\n'
        if feature.ndim == 2:  # Flattened input
            # str=str+'Reshaping 2 dim\n'
            feature = np.reshape(feature, (sequences, timesteps, strides * bands))  # was frames x bands*timesteps
            if truth is not None:
                truth = np.reshape(truth, (sequences, timesteps, num_classes))  # was frames x num_classes
        elif feature.ndim == 3:  # Unflattened input
            # str=str+'Reshaping 3 dim\n'
            feature = np.reshape(feature, (sequences, timesteps, strides, bands))  # was frames x bands x timesteps
            if truth is not None:
                truth = np.reshape(truth, (sequences, timesteps, num_classes))  # was frames x num_classes
    else:
        # Drop frames if remainder exists (not fitting into a multiple of new number of sequences)
        fr2drop = feature.shape[0] % batch_size
        if fr2drop > 0:
            msg += f'Dropping {fr2drop} frames for total to be a multiple of batch_size\n'
            feature = feature[0:-fr2drop, ]
            if truth is not None:
                truth = truth[0:-fr2drop, ]

    # Add channel dimension if required for input to model (i.e. for cnn type input)
    if add1ch:
        msg += 'Adding channel dimension to feature\n'
        feature = np.expand_dims(feature, axis=feature.ndim)  # add as last/outermost dim

    in_shape = feature.shape
    in_shape = in_shape[1:]  # remove frame dim size

    msg += f'Feature final shape: {feature.shape}\n'
    msg += f'Input shape final (includes timesteps): {in_shape}\n'
    if truth is not None:
        msg += f'Truth final shape: {truth.shape}\n'

    return feature, truth, in_shape, num_classes, class_weights, msg


def reshape_outputs(predict: np.ndarray,
                    truth: np.ndarray = None,
                    timesteps: int = 0) -> (np.ndarray, np.ndarray, int):
    """Reshape model output data.

    truth and predict can be either frames x num_classes, or frames x timesteps x num_classes
    In binary case, num_classes dim may not exist; detect this and set num_classes to 1.
    """
    from sonusai import SonusAIError

    if truth is not None:
        if predict.shape != truth.shape:
            raise SonusAIError('predict and truth shapes do not match')

    num_dims = predict.ndim
    dims = predict.shape

    if num_dims == 3 or (num_dims == 2 and timesteps > 0):
        if num_dims == 2:
            # 2D with timesteps - frames x timesteps
            num_classes = 1
        else:
            # 3D - frames x timesteps x num_classes
            num_classes = dims[2]

        # reshape to remove timestep dimension
        new_dims = (dims[0] * dims[1], num_classes)
        predict = np.reshape(predict, new_dims)
        if truth is not None:
            truth = np.reshape(truth, new_dims)
    else:
        if num_dims == 1:
            # 1D without timesteps - frames
            num_classes = 1

            # convert to 2D - frames x 1
            predict = np.expand_dims(predict, 1)
            if truth is not None:
                truth = np.expand_dims(truth, 1)
        else:
            # 2D without timesteps - frames x num_classes
            num_classes = dims[1]

    return predict, truth, num_classes
